@@ -20,18 +20,19 @@
 
 import copy
 import errno
+import ipaddress
 import logging
 import os
 import re
+from urllib import parse as urlparse
 
 from oslo_concurrency import processutils
 from oslo_config import cfg
+from oslo_service import loopingcall
 from oslo_utils import excutils
 from oslo_utils import specs_matcher
 from oslo_utils import strutils
 from oslo_utils import units
-import six
-from six.moves.urllib import parse
 
 from ironic_lib.common.i18n import _
 from ironic_lib import exception
@@ -193,7 +194,7 @@ def _extract_hint_operator_and_values(hint_expression, hint_name):
         :op: The operator. An empty string in case of None.
         :values: A list of values stripped and converted to lowercase.
     """
-    expression = six.text_type(hint_expression).strip().lower()
+    expression = str(hint_expression).strip().lower()
     if not expression:
         raise ValueError(
             _('Root device hint "%s" expression is empty') % hint_name)
@@ -226,7 +227,7 @@ def _normalize_hint_expression(hint_expression, hint_name):
     :returns: A normalized string.
     """
     hdict = _extract_hint_operator_and_values(hint_expression, hint_name)
-    result = hdict['op'].join([' %s ' % parse.quote(t)
+    result = hdict['op'].join([' %s ' % urlparse.quote(t)
                                for t in hdict['values']])
     return (hdict['op'] + result).strip()
 
@@ -247,7 +248,7 @@ def _append_operator_to_hints(root_device):
         if VALID_ROOT_DEVICE_HINTS[name] is bool:
             continue
 
-        expression = six.text_type(expression)
+        expression = str(expression)
         ast = ROOT_DEVICE_HINTS_GRAMMAR.parseString(expression)
         if len(ast) > 1:
             continue
@@ -288,7 +289,7 @@ def parse_root_device_hints(root_device):
     for name, expression in root_device.items():
         hint_type = VALID_ROOT_DEVICE_HINTS[name]
         if hint_type is str:
-            if not isinstance(expression, six.string_types):
+            if not isinstance(expression, str):
                 raise ValueError(
                     _('Root device hint "%(name)s" is not a string value. '
                       'Hint expression: %(expression)s') %
@@ -325,6 +326,101 @@ def parse_root_device_hints(root_device):
     return _append_operator_to_hints(root_device)
 
 
+def find_devices_by_hints(devices, root_device_hints):
+    """Find all devices that match the root device hints.
+
+    Try to find devices that match the root device hints. In order
+    for a device to be matched it needs to satisfy all the given hints.
+
+    :param devices: A list of dictionaries representing the devices
+                    containing one or more of the following keys:
+
+        :name: (String) The device name, e.g /dev/sda
+        :size: (Integer) Size of the device in *bytes*
+        :model: (String) Device model
+        :vendor: (String) Device vendor name
+        :serial: (String) Device serial number
+        :wwn: (String) Unique storage identifier
+        :wwn_with_extension: (String): Unique storage identifier with
+                             the vendor extension appended
+        :wwn_vendor_extension: (String): United vendor storage identifier
+        :rotational: (Boolean) Whether it's a rotational device or
+                     not. Useful to distinguish HDDs (rotational) and SSDs
+                     (not rotational).
+        :hctl: (String): The SCSI address: Host, channel, target and lun.
+                         For example: '1:0:0:0'.
+        :by_path: (String): The alternative device name,
+                  e.g. /dev/disk/by-path/pci-0000:00
+
+    :param root_device_hints: A dictionary with the root device hints.
+    :raises: ValueError, if some information is invalid.
+    :returns: A generator with all matching devices as dictionaries.
+    """
+    LOG.debug('Trying to find devices from "%(devs)s" that match the '
+              'device hints "%(hints)s"',
+              {'devs': ', '.join([d.get('name') for d in devices]),
+               'hints': root_device_hints})
+    parsed_hints = parse_root_device_hints(root_device_hints)
+    for dev in devices:
+        device_name = dev.get('name')
+
+        for hint in parsed_hints:
+            hint_type = VALID_ROOT_DEVICE_HINTS[hint]
+            device_value = dev.get(hint)
+            hint_value = parsed_hints[hint]
+
+            if hint_type is str:
+                try:
+                    device_value = _normalize_hint_expression(device_value,
+                                                              hint)
+                except ValueError:
+                    LOG.warning(
+                        'The attribute "%(attr)s" of the device "%(dev)s" '
+                        'has an empty value. Skipping device.',
+                        {'attr': hint, 'dev': device_name})
+                    break
+
+            if hint == 'size':
+                # Since we don't support units yet we expect the size
+                # in GiB for now
+                device_value = device_value / units.Gi
+
+            LOG.debug('Trying to match the device hint "%(hint)s" '
+                      'with a value of "%(hint_value)s" against the same '
+                      'device\'s (%(dev)s) attribute with a value of '
+                      '"%(dev_value)s"', {'hint': hint, 'dev': device_name,
+                                          'hint_value': hint_value,
+                                          'dev_value': device_value})
+
+            # NOTE(lucasagomes): Boolean hints are not supported by
+            # specs_matcher.match(), so we need to do the comparison
+            # ourselves
+            if hint_type is bool:
+                try:
+                    device_value = strutils.bool_from_string(device_value,
+                                                             strict=True)
+                except ValueError:
+                    LOG.warning('The attribute "%(attr)s" (with value '
+                                '"%(value)s") of device "%(dev)s" is not '
+                                'a valid Boolean. Skipping device.',
+                                {'attr': hint, 'value': device_value,
+                                 'dev': device_name})
+                    break
+                if device_value == hint_value:
+                    continue
+
+            elif specs_matcher.match(device_value, hint_value):
+                continue
+
+            LOG.debug('The attribute "%(attr)s" (with value "%(value)s") '
+                      'of device "%(dev)s" does not match the hint %(hint)s',
+                      {'attr': hint, 'value': device_value,
+                       'dev': device_name, 'hint': hint_value})
+            break
+        else:
+            yield dev
+
+
 def match_root_device_hints(devices, root_device_hints):
     """Try to find a device that matches the root device hints.
 
@@ -355,65 +451,119 @@ def match_root_device_hints(devices, root_device_hints):
     :raises: ValueError, if some information is invalid.
     :returns: The first device to match all the hints or None.
     """
-    LOG.debug('Trying to find a device from "%(devs)s" that matches the '
-              'root device hints "%(hints)s"',
-              {'devs': ', '.join([d.get('name') for d in devices]),
-               'hints': root_device_hints})
-    parsed_hints = parse_root_device_hints(root_device_hints)
-    for dev in devices:
-        device_name = dev.get('name')
+    try:
+        dev = next(find_devices_by_hints(devices, root_device_hints))
+    except StopIteration:
+        LOG.warning('No device found that matches the root device hints %s',
+                    root_device_hints)
+    else:
+        LOG.info('Root device found! The device "%s" matches the root '
+                 'device hints %s', dev, root_device_hints)
+        return dev
 
-        for hint in parsed_hints:
-            hint_type = VALID_ROOT_DEVICE_HINTS[hint]
-            device_value = dev.get(hint)
-            hint_value = parsed_hints[hint]
 
-            if hint_type is str:
-                try:
-                    device_value = _normalize_hint_expression(device_value,
-                                                              hint)
-                except ValueError:
-                    LOG.warning(
-                        'The attribute "%(attr)s" of the device "%(dev)s" '
-                        'has an empty value. Skipping device.',
-                        {'attr': hint, 'dev': device_name})
-                    break
+def wait_for_disk_to_become_available(device):
+    """Wait for a disk device to become available.
 
-            if hint == 'size':
-                # Since we don't support units yet we expect the size
-                # in GiB for now
-                device_value = device_value / units.Gi
+    Waits for a disk device to become available for use by
+    waiting until all process locks on the device have been
+    released.
 
-            LOG.debug('Trying to match the root device hint "%(hint)s" '
-                      'with a value of "%(hint_value)s" against the same '
-                      'device\'s (%(dev)s) attribute with a value of '
-                      '"%(dev_value)s"', {'hint': hint, 'dev': device_name,
-                                          'hint_value': hint_value,
-                                          'dev_value': device_value})
+    Timeout and iteration settings come from the configuration
+    options used by the in-library disk_partitioner:
+    ``check_device_interval`` and ``check_device_max_retries``.
 
-            # NOTE(lucasagomes): Boolean hints are not supported by
-            # specs_matcher.match(), so we need to do the comparison
-            # ourselves
-            if hint_type is bool:
-                try:
-                    device_value = strutils.bool_from_string(device_value,
-                                                             strict=True)
-                except ValueError:
-                    LOG.warning('The attribute "%(attr)s" (with value '
-                                '"%(value)s") of device "%(dev)s" is not '
-                                'a valid Boolean. Skipping device.',
-                                {'attr': hint, 'value': device_value,
-                                 'dev': device_name})
-                    break
-                if device_value == hint_value:
-                    continue
-                break
+    :params device: The path to the device.
+    :raises: IronicException If the disk fails to become
+        available.
+    """
+    retries = [0]
+    pids = ['']
+    stderr = ['']
+    interval = CONF.disk_partitioner.check_device_interval
+    max_retries = CONF.disk_partitioner.check_device_max_retries
 
-            if not specs_matcher.match(device_value, hint_value):
-                break
+    def _wait_for_disk(device, retries, max_retries, pids, stderr):
+        # A regex is likely overkill here, but variations in fuser
+        # means we should likely use it.
+        fuser_pids_re = re.compile(r'\d+')
+
+        retries[0] += 1
+        if retries[0] > max_retries:
+            raise loopingcall.LoopingCallDone()
+
+        # There are 'psmisc' and 'busybox' versions of the 'fuser' program. The
+        # 'fuser' programs differ in how they output data to stderr.  The
+        # busybox version does not output the filename to stderr, while the
+        # standard 'psmisc' version does output the filename to stderr.  How
+        # they output to stdout is almost identical in that only the PIDs are
+        # output to stdout, with the 'psmisc' version adding a leading space
+        # character to the list of PIDs.
+        try:
+            # NOTE(ifarkas): fuser returns a non-zero return code if none of
+            #                the specified files is accessed.
+            # NOTE(TheJulia): fuser does not report LVM devices as in use
+            #                 unless the LVM device-mapper device is the
+            #                 device that is directly polled.
+            # NOTE(TheJulia): The -m flag allows fuser to reveal data about
+            #                 mounted filesystems, which should be considered
+            #                 busy/locked. That being said, it is not used
+            #                 because busybox fuser has a different behavior.
+            # NOTE(TheJuia): fuser outputs a list of found PIDs to stdout.
+            #                All other text is returned via stderr, and the
+            #                output to a terminal is merged as a result.
+            out, err = execute('fuser', device, check_exit_code=[0, 1],
+                               run_as_root=True)
+
+            if not out and not err:
+                raise loopingcall.LoopingCallDone()
+
+            stderr[0] = err
+            # NOTE: findall() returns a list of matches, or an empty list if no
+            # matches
+            pids[0] = fuser_pids_re.findall(out)
+
+        except processutils.ProcessExecutionError as exc:
+            LOG.warning('Failed to check the device %(device)s with fuser:'
+                        ' %(err)s', {'device': device, 'err': exc})
+
+    timer = loopingcall.FixedIntervalLoopingCall(
+        _wait_for_disk,
+        device, retries, max_retries, pids, stderr)
+    timer.start(interval=interval).wait()
+
+    if retries[0] > max_retries:
+        if pids[0]:
+            raise exception.IronicException(
+                _('Processes with the following PIDs are holding '
+                  'device %(device)s: %(pids)s. '
+                  'Timed out waiting for completion.')
+                % {'device': device, 'pids': ', '.join(pids[0])})
         else:
-            LOG.info('Device found! The device "%s" matches the root '
-                     'device hints', device_name)
-            return dev
+            raise exception.IronicException(
+                _('Fuser exited with "%(fuser_err)s" while checking '
+                  'locks for device %(device)s. Timed out waiting for '
+                  'completion.')
+                % {'device': device, 'fuser_err': stderr[0]})
 
-    LOG.warning('No device found that matches the root device hints')
+
+def get_route_source(dest, ignore_link_local=True):
+    """Get the IP address to send packages to destination."""
+    try:
+        out, _err = execute('ip', 'route', 'get', dest)
+    except (EnvironmentError, processutils.ProcessExecutionError) as e:
+        LOG.warning('Cannot get route to host %(dest)s: %(err)s',
+                    {'dest': dest, 'err': e})
+        return
+
+    try:
+        source = out.strip().split('\n')[0].split('src')[1].split()[0]
+        if (ipaddress.ip_address(source).is_link_local
+                and ignore_link_local):
+            LOG.debug('Ignoring link-local source to %(dest)s: %(rec)s',
+                      {'dest': dest, 'rec': out})
+            return
+        return source
+    except (IndexError, ValueError):
+        LOG.debug('No route to host %(dest)s, route record: %(rec)s',
+                  {'dest': dest, 'rec': out})
